/* This file was generated by syncqt. */
#ifndef QT_QTWAYLANDCLIENT_VERSION_H
#define QT_QTWAYLANDCLIENT_VERSION_H

#define QTWAYLANDCLIENT_VERSION_STR "5.12.8"

#define QTWAYLANDCLIENT_VERSION 0x050C08

#endif // QT_QTWAYLANDCLIENT_VERSION_H
